""" Adds additional language support to the testing library
"""
from typing import List
import os
import subprocess

def c_compile(c_files: List[str], out_file:str = "a.out", force:bool = False, compiler="clang -Wall") -> str:
    """Compiles using the clang compiler by default. Will only compile if 
    the file doesn't already exist

    Args:
        c_files (List[str]): list of input files
        out_file (str, optional): _the file to save out. Defaults to "a.out".
        force (bool, optional): Force recompile even if it already exists. Defaults to False.
        compiler (str, optional): the compiler to use with arguments. Defaults to clang -Wall
    Returns:
        str: A list of warnings and errors if they happen.
    """
    if not os.path.isfile(out_file) or force:
        compile_command = compiler.split() + c_files + ["-o", out_file]
        compiled = subprocess.run(compile_command, capture_output=True)
        if (compiled.stderr):
            return "Code failed to compile or had warnings\n{0}".format(compiled.stderr.decode())
    return ''


def run_range(file:str, start : int , end: int) -> str:
        """Runs subprocess passing in start and end as the first two parameters 
        
        Args:
            file (str): the file / program to run
            start (int): test id of the start test
            end (int): test id of the end test
        Returns:
            str: items printed to stderr
        """
        command = subprocess.run([file, str(start), str(end)], capture_output=True)
        if (command.stderr):
            return command.stderr.decode()
        return ''


"""
Support script for building autograders.
"""
import click
from zipfile import ZipFile
import os
from typing import List, Union, Tuple, Dict, Any
import neugs_utils.scripts.templates as templates


def get_config_files(config: Dict[str, Any]) -> Dict[str, str]:
    """Based on config settings, builds the four
    autograder required files:
        requirements.txt
        run_autograder
        run_tests.py
        setup.sh

    Args:
        config (Dict[str, Any]): config settings

    Returns:
        Dict[str, str]: file name, and contents.
    """
    config_files = {}
    config_files["requirements.txt"] = templates.get_template("requirements.txt")
    config_files["run_autograder"] = templates.get_template("run_autograder")
    config_files["setup.sh"] = templates.build_setup(config)
    config_files["run_tests.py"] = templates.build_runner(config)
    return config_files


def get_all_file_paths(directory: str) -> List[str]:
    """
        Grabs all the files in the directory provided, to write them out to the tests directory.
        While it will get files in sub directories, it does not currently handle
        keeping those subdirectories when writing!
    Args:
        directory (str): directory to scan

    Returns:
        List[str]: a list of files found
    """
    # initializing empty file paths list
    file_paths = []

    # crawling through directory and subdirectories
    for root, directories, files in os.walk(directory):
        for filename in files:
            if "__" in root:
                continue  # skip special python directories TODO: improve this!
            # join the two strings in order to form the full filepath.
            filepath = os.path.join(root, filename)
            file_paths.append(filepath)

    # returning all file paths
    return file_paths


def new_test_loc(file: str) -> str:
    """converts a file found to be put into the tests directory
    TODO: this function should be updated to be more flexible.
    Args:
        file (str): the file to convert path to tests directory in the zip

    Returns:
        str: tests/file
    """
    return "tests/" + os.path.basename(file)


@click.command()
@click.option(
    "--out",
    "-o",
    default="autograder.zip",
    show_default=True,
    help="The file name of the autograder.",
)
@click.option(
    "--test-dir",
    "-d",
    default="tests",
    show_default=True,
    help="The directory to pull tests from.",
)
@click.option(
    "--cpp", default=False, is_flag=True, help="Add C/CPP compiler to setup files."
)
@click.option(
    "--grading-type",
    "-t",
    default="mastery",
    show_default=True,
    help="Grading types allowed [mastery, standard]",
)
@click.option(
    "--test-only",
    is_flag=True,
    default=False,
    help="Print out feedback, but don't build zip.",
)
def build_autograder(
    out: str, test_dir: str, cpp: bool, grading_type: str, test_only: bool
) -> None:
    """Command line tool for generating autograder_files. For the most part
    the default execution will work. Reads files in dir, addes them
    to a zipfile (out), and can add some of the standard four autograder files
    incase they want to be replaced
    Args:
        out (str): file to save
        dir (str): tests directory
        cpp (bool): include c/cpp compilers
        grading_type (str): mastery or standard
        test_only (bool): run system but don't write to zip file
    """
    paths = get_all_file_paths(test_dir)

    config = {"cpp": cpp, "gtype": grading_type.casefold()}

    with ZipFile(out, "w") as zip:
        for file in paths:
            if test_only:
                print(f"Testing: would be writing {file} to zip.")
            else:
                zip.write(file, arcname=new_test_loc(file))
        for file, contents in get_config_files(config).items():
            if test_only:
                print(f"...writing to zip {file} with \n{contents}")
            else:
                zip.writestr(data=contents, zinfo_or_arcname=file)



if __name__ == "__main__":
    build_autograder()



from typing import Dict, Any


__setup_template = """#!/usr/bin/env bash
## Autogenerated template by neugs_utils
{c_setup}
apt-get install -y python3 python3-pip python3-dev
pip3 install -r /autograder/source/requirements.txt
"""

__c_setup = """
apt-get install -y clang
apt-get install -y gcc
"""


def build_setup(config: Dict[str, Any]) -> str:
    """Builds the setup based on config options. Currently
    just simple subs


    Args:
        config (Dict[str, Any]): the config options

    Returns:
        str: the file to write
    """
    cstring = __c_setup if config["cpp"] else ""
    return __setup_template.format(c_setup=cstring)



"""
Support script for adding tests
"""
import click
from zipfile import ZipFile
import os
from typing import List

_common_header = """
import unittest
from gradescope_utils.autograder_utils.decorators import number
from neugs_utils.context_managers import Capturing
from neugs_utils import tier, COMMON_ONE, COMMON_TWO, COMMON_THREE, common_msg, strip_prompts
from neugs_utils.common_tests import check_style


student_files = [] # setup incase there are more than one import section
"""


_common_main = """
if __name__ == '__main__':
    unittest.main()
"""

_import_template = """
## Autogenerated import based on test-add
try:
    import {0}
    student_files += ['{1}']
except ImportError:
    try:
        import {2} as {0}
        student_files += ['{3}']
    except ImportError:
        pass

"""

#class TestStarRating(unittest.TestCase):


def build_import(file : str) -> str: 
    """builds a series of try/except statements based on _import_template.

    Args:
        file (str): the file to import

    Returns:
        str: returns an import statement to work locally 
        and on gs auto-grader
    """
    # first get the file parts
    file = os.path.normpath(file).replace("\\", "/")
    dir = os.path.dirname(file)
    file_only = os.path.basename(file)
    mod_base = os.path.splitext(file_only)[0]
    mod_long = '.'.join(dir.split('/')) + "." + mod_base
    return _import_template.format(mod_base, file_only, mod_long, file)




@click.command()
@click.option("--cl", "-c", default="Test")
@click.argument("out")
@click.argument("modules", nargs=-1)
def new_test(out : str,  cl : str, modules : List[str]) -> None:
    """Builds a new test file. Requires the file to write to,
    it is suggest to added a class, but not required, and
    suggested to add modules which match the student files 
    being tested. Currently defaults to
    a set of imports related to TierMastery grading. 

    Args:
        out (str): file to save to
        cl (str): test class
        modules (List[str]): files to import (usually student files)
    """
    imports = ''
    for file in modules:
        imports += build_import(file)
    contents = _common_header
    contents += imports
    contents += "\n" ## add an extra blank
    contents += "# Double check class name, remember unit tests likes it to start with Test\n"
    contents += f"class {cl}(unittest.TestCase):\n    pass\n\n"

    contents += _common_main
    with open(out, 'w') as f:
        f.write(contents)

if __name__ == "__main__":
    new_test()



## Autogenerated template by neugs_utils

import unittest
from neugs_utils import TierMasteryJSONTestRunner

if __name__ == '__main__':
    suite = unittest.defaultTestLoader.discover('tests')
    with open('/autograder/results/results.json', 'w') as f:
        TierMasteryJSONTestRunner(visibility='visible', stream=f).run(suite)



from io import StringIO 
import sys

#inspired by https://stackoverflow.com/questions/16571150/how-to-capture-stdout-output-from-a-python-function-call
class Capturing(list):
    """Context capture class, that focuses on capturing output and allowing
    for StringIO input. 

    output is a list is list that contains every line of output as an item.

    Example:
        with Capturing("hello") as output:
            some_func()
        self.assertEqual(output[0], "hello")
    """
    def __init__(self, input = None):
        """Setups the context. Allows for an optional input string. 

        Args:
            input (str or StringIO(), optional): Can be a string, or StringIO(), 
            replaces sys.in context. Defaults to None which means sys.in is not modified.
        """
        if input is not None:
            if type(input) == str:
                input =StringIO(input)
            self._input = input
        else:
            self._input = None

    def __enter__(self):
        self._stdout = sys.stdout
        if self._input is not None:
            self._stdin = sys.stdin
            sys.stdin = self._input
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout
        if self._input is not None:
            sys.stdin = self._stdin



from gradescope_utils.autograder_utils.files import check_submitted_files
import io
import contextlib
import pycodestyle
from typing import List


# Developed by Mark Miller for Khoury Align SF campus use. Placed for easy access.

common_style_ignore_list = [
    'E121', # Continuation line under-indented for hanging indent
    'E126', # Continuation line over-indented for hanging indent
    'W291', # Trailing whitespace
    'W503', # Line break before binary operator
    'W504', # Line break after binary operator
    'E501' # Relaxing the line length due to comments commonly going over 80
    ]


def check_files(student_files : List[str]) -> str:
    """checks to see if the files exist in the submission

    Args:
        student_files (List[str]): the names of files to check for

    Returns:
        str: A string listing which files were missing, or the empty string if no files were missing.
    """
    missing_files = check_submitted_files(student_files)
    return "Missing required files:\n\t" + " ".join(missing_files) if len(missing_files) != 0 else ''


def check_style(student_files: List[str], ignore : List[str] = common_style_ignore_list ) -> str:
    """Runs the files against a style checker to make sure they match the recommended python styles. 

    Args:
        student_files (List[str]): The student files to run pycodestyle on
        ignore (List[str], ignore, optional): List of ignores to pass onto pycodestyle . Defaults to common_style_ignore_list

    Returns:
        str:  string  if any style did not pass the style checker, or empty string if everything passed fine 
    """
    with io.StringIO() as buf, contextlib.redirect_stdout(buf):
        style_checker = pycodestyle.StyleGuide(ignore=ignore)
        report = style_checker.check_files(student_files)        
        return "\nDoes not pass the style checker:\n" + buf.getvalue() if report.total_errors != 0 else ''



import os
from .setup_file import build_setup
from .test_runner import build_runner


def get_template(filename: str) -> str:
    """Loads a template file into a string

    Args:
        filename (str): the file to load

    Returns:
        str: the file as a string
    """
    dirname = os.path.dirname(__file__)
    file = os.path.join(dirname, filename)

    f = open(file)
    contents = f.read()
    f.close()
    return contents



from typing import Dict, Any


__mastery = """
## Autogenerated runner by neugs_utils
import unittest
from neugs_utils import TierMasteryJSONTestRunner

if __name__ == '__main__':
    suite = unittest.defaultTestLoader.discover('tests')
    with open('/autograder/results/results.json', 'w') as f:
        TierMasteryJSONTestRunner(visibility='visible', stream=f).run(suite)

"""


__standard = """
## Autogenerated runner by neugs_utils
import unittest
from gradescope_utils.autograder_utils.json_test_runner import JSONTestRunner

if __name__ == '__main__':
    suite = unittest.defaultTestLoader.discover('tests')
    with open('/autograder/results/results.json', 'w') as f:
        JSONTestRunner(visibility='visible', stream=f).run(suite)

"""

def build_runner(config: Dict[str, Any]) -> str:
    """build a runner based on the type of grading system to use

    Args:
        config (Dict[str, Any]): a config file

    Returns:
        str: the file to write out
    """
    return __mastery if config["gtype"] == "mastery" else __standard



import unittest
import sys
import os
from neugs_utils import TierMasteryJSONTestRunner

if __name__ == '__main__':
    #up1 = os.path.abspath('..')
    #print("Add path for test purposes " + up1)
    #sys.path.insert(0, up1)
   
    suite = unittest.defaultTestLoader.discover('tests/examples')
    with open('results.json', 'w') as f:
        TierMasteryJSONTestRunner(visibility='visible', stream=f).run(suite)
